from . import sql
import os
import re

class SQLMap:
	def __init__ (self, map = None, auto_reload = False, engine = "postgresql", conn = None, sqls = None):
		self._map = map
		self._auto_reload = auto_reload
		self._engine = engine
		self._conn = conn
		self._version = "1.0"
		self._sqls = sqls or {}
		self._last_modifed = 0
		if self._map:
			self._read_from_file ()
	
	def __enter__ (self):
		return self

	def __exit__ (self, type, value, tb):
		pass
	
	def __getattr__ (self, name):		
		self._reloaderble () and self._read_from_file ()
		if self._sqls.get (name) is None:
			raise AttributeError ("SQL template '{}' not found".format (name))
		return sql.SQLTemplateRenderer (self._sqls.get (name), self._engine, self._conn)
		
	def _reloaderble (self):
		return self._map and self._auto_reload and self._last_modifed != os.path.getmtime (self._map)
	
	def __resolve (self, table, alias):
		return alias and (table + " AS " + alias) or table
	
	def new (self, conn):	
		return SQLMap (None, self._auto_reload, conn.dbtype, conn, sqls = self._sqls)
	
	def execute (self, *args, **kargs):
		return self._conn.execute (*args, **kargs)

	def commit (self, *args, **kargs):
		return self._conn.commit (*args, **kargs)

	def rollback (self, *args, **kargs):
		return self._conn.rollback (*args, **kargs)

	def template (self, template):
		return sql.SQLTemplateRenderer (template, self._engine, self._conn)
	from_string = template
	
	def _render_table (self, table):
		if isinstance (table, str):
			return table
		return "({})".format (table)
			
	def insert (self, table, alias = None):
		return sql.SQLComposer ("INSERT INTO " + self.__resolve (self._render_table (table), alias) + " ({this.columns}) VALUES ({this.values})", self._engine, self._conn)
		
	def update (self, table, alias = None):		
		return sql.SQLComposer ("UPDATE " + self.__resolve (self._render_table (table), alias) + " SET {this.pairs}", self._engine, self._conn, True)
		
	def select (self, table, alias = None):
		return sql.SQLComposer ("SELECT {select} FROM " + self.__resolve (self._render_table (table), alias), self._engine, self._conn).feed (select = "*")
		
	def delete (self, table, alias = None):
		return sql.SQLComposer ("DELETE FROM " + self.__resolve (self._render_table (table), alias), self._engine, self._conn, True)
		
	def _read_from_file (self):
		self._last_modifed = os.path.getmtime (self._map)
		with open (self._map) as f:
			self._read_from_string (f.read ())
	
	RX_NAME	= re.compile ("\sname\s*=\s*['\"](.+?)['\"]")
	RX_VERSION	= re.compile ("\sversion\s*=\s*['\"](.+?)['\"]")
	def _read_from_string (self, data):
		self._sqls = {}
		
		current_name = None
		current_data = []
		for line in data.split ("\n"):
			if not line.strip ():
				continue
			
			if line.startswith ("<sqlmap "):
				m = self.RX_VERSION.search (line)
				if m:					
					self._version = m.group (1)
				
			elif line.startswith ("</sql>"):
				if not current_name:
					raise ValueError ("unexpected end tag </sql>")
				self._sqls [current_name] = "\n".join (current_data)
				current_name, current_data = None, []				 
			
			elif line.startswith ("<sql "):
				m = self.RX_NAME.search (line)
				if not m:
					raise ValueError ("name attribute required")
				current_name = m.group (1)
			
			elif current_name:
				current_data.append (line.strip ())




from . import db3
from .dbtypes import DB_PGSQL, DB_SQLITE3
import sqlphile

try:
    import psycopg2

except ImportError:
    class open:
        def __init__ (self, *args, **kargs):
            raise ImportError ('psycopg2 not installed')
    open3 = open2 = open

else:
    class open (db3.open):
        def __init__ (self, dbname, user, password, host = '127.0.0.1', port = 5432, dir = None, auto_reload = False, auto_closing = True):
            self.closed = False
            self.auto_closing = auto_closing

            self.conn = None
            if ":" in host:
                host, port = host.split (":")
                port = int (port)
            self.conn = psycopg2.connect (host=host, dbname=dbname, user=user, password=password, port = port)
            self._init (dir, auto_reload, DB_PGSQL)

        def field_names (self):
            return [x.name for x in self.description]

        def set_autocommit (self, flag = True):
            self.conn.autocommit = flag


    # -----------------------------------------------------------------

    class open2 (open, db3.open2):
        def __init__ (self, conn, dir = None, auto_reload = False, auto_closing = True):
            self.closed = False
            self.conn = conn
            self.auto_closing = auto_closing
            self._init (dir, auto_reload, DB_PGSQL)

    # -----------------------------------------------------------------

    class open3 (db3.open3):
        def __init__ (self, conn, dir = None, auto_reload = False, auto_closing = True):
            self.closed = False
            self.conn = conn
            self.auto_closing = False
            self._init (dir, auto_reload, DB_PGSQL)



import os
import re
from . import sql
from .sql import SQL, SQLTemplateRenderer, SQLComposer
from .sqlmap import SQLMap
from .q import Q, V
from .d import D
from .f import F
from . import db3, pg2
from .utils import *
from .dbtypes import DB_PGSQL, DB_SQLITE3, DB_SYN_PGSQL

__version__ = "0.6.3.6"

sqlite3 = db3.open
postgres = pg2.open

class SQLPhile (SQLMap):
	def __init__ (self, dir = None, auto_reload = False, engine = DB_PGSQL, conn = None):
		self._dir = dir
		self._auto_reload = auto_reload
		self._engine = engine
		self._conn = conn
		self._ns = {}
		self._dir and self._load_sqlmaps ()

	def __getattr__ (self, name):
		try:
			return self._ns [name]
		except KeyError:
			return getattr (self._ns ["default"], name)

	def _load_sqlmaps (self):
		for fn in os.listdir (self._dir):
			if fn[0] == "#":
				continue
			ns = fn.split (".") [0]
			if ns == "ops":
				raise NameError ('ops cannot be used SQL map file name')
			self._ns [ns] = SQLMap (os.path.join (self._dir, fn), self._auto_reload, self._engine, self._conn)


def Template (backend = DB_PGSQL, path = None, auto_reload = False):
	if backend [0] != "*":
		backend = "*" + backend
	assert backend in (DB_PGSQL, DB_SQLITE3, DB_SYN_PGSQL)
	if path and os.path.isdir (path):
		return SQLPhile (path, auto_reload, backend)
	return SQLMap (path, auto_reload, backend)



try:
    from skitai.corequest.httpbase.task import Result

except ImportError:
    EMPTY = None
    dispatch = None

else:
    class Response:
        def __init__ (self, data, expt):
            self.description = None
            self.data = data
            self.expt = expt
            self.code, self.msg = 200, "OK"
            self.status_code, self.reason = self.code, self.msg  

    def dispatch (data, expt = None):
        return Result (None, 3, Response (data, expt), None)

    EMPTY = dispatch ([])    



from . import utils
from .f import F
from . import d
from .dbtypes import DB_PGSQL, DB_SQLITE3

OPS = {
	"gt": ">",
	"gte": ">=",
	"lt": "<",
	"lte": "<=",
	"neq": "<>",
	"eq": "=",
	"exact": "=",
	"regex": "~",
	"in": "IN",
	"notin": "NOT IN",
	"isnull": "IS",
	"exists": "EXISTS",
	"nexists": "NOT EXISTS",
	"between": "BETWEEN",
	"contains": "LIKE",
	"startswith": "LIKE",
	"endswith": "LIKE",
	"ncontains": "NOT LIKE",
	"nstartswith": "NOT LIKE",
	"nendswith": "NOT LIKE",
	"all": "=",
}

class _Q:
	def __init__ (self, *args, **kargs):
		self._clause = None
		self._value = None
		self._void = False

		if kargs:
			assert len (kargs) == 1
			self.k, self.v = kargs.popitem ()
		elif len (args) == 0:
			self.k, self.v = "", None
		elif len (args) == 2:
			self.k, self.v = args
		else:
			self._clause = args [0]
		self._exclude = False

	def add_percent (self, val, pos = 0):
		if isinstance (val, F):
			val.add_percent (pos)
			return val
		val = val.replace ("%", "\\%")
		if pos == 0:
			return "%" + val + "%"
		elif pos == 1:
			return "%" + val
		return val + "%"

	def render (self, engine = DB_PGSQL):
		from .sql import SQL

		if self._clause is not None:
			if self._exclude and self._clause:
				self._clause = "NOT (" + self._clause + ")"
				self._exclude = False
			return self._clause

		elif self.v is None:
			self._clause = ""
			self._void = True
			return self._clause

		else:
			k, v = self.k, self.v
			ll = k.split ("__")
			if len (ll) == 1:
				fd, op = k, "eq"
			else:
				if ll [-1] in (OPS):
					fd, op = ".".join (ll [:-1]), ll [-1]
				else:
					fd, op = ".".join (ll), "eq"

			try:
				_op = OPS [op]
			except KeyError:
				raise TypeError ('Unknown Operator: {}'.format (op))

			_val = v
			_escape = True
			if op.endswith ("all"):
				fd, _val = "1", int (_val)
			elif op.endswith ("contains"):
				_val = self.add_percent (_val, 0)
			elif op.endswith ("endswith"):
				_val = self.add_percent (_val, 1)
			elif op.endswith ("startswith"):
				_val = self.add_percent (_val, 2)
			elif op == "between":
				_val = "{} AND {}".format (d.toval (_val [0], engine), d.toval (_val [1], engine))
				_escape = False
			elif op == "isnull":
				if not _val: # False
					_op = "IS NOT"
				_val = None
			elif op == "in":
				if isinstance (_val, (list, tuple)):
					_val = "({})".format (",".join ([d.toval (each, engine) for each in _val]))
				else:
					_val = "({})".format (str (_val))
				_escape = False
			elif isinstance (_val, SQL):
				_val = "({})".format (str (_val))
				_escape = False

			self._value = _escape and d.toval (_val, engine) or _val
			self._clause = "{} {} {}".format (fd, _op, self._value)

		if self._exclude and self._clause:
			self._clause =  "NOT (" + self._clause + ")"
			self._exclude = False
		return self._clause

	def _joinwith (self, op, b):
		if self and b:
			return Q ("({} {} {})".format (self, op, b))
		return self or b

	def __len__ (self):
		return (self._clause or self.v is not None) and 1 or 0

	def __str__ (self):
		return self.render ()

	def __repr__ (self):
		return "<sqlphile.q.Q_ object: {}>".format (self.render ())

	def __or__ (self, b):
		return self._joinwith ("OR", b)

	def __and__ (self, b):
		return self._joinwith ("AND", b)

	def __invert__ (self):
		self._exclude = True
		return self

def Q (*args, **kargs):
	clauses = []
	if kargs:
		assert len (kargs)
		for k, v in kargs.items ():
			clauses.append (_Q (**{k : v}))

		if len (clauses) == 1:
			return clauses [0]
		else:
			stack = []
			for clause in clauses:
				res = clause.render ()
				if res:
					stack.append (res)
			if stack:
				return _Q ("(" + " AND ".join (stack) + ")")
			else:
				return _Q ("")
	return _Q (*args)

class V (_Q):
	def __init__ (self, *args, **kargs):
		self._clause = None
		self._value = None
		if kargs:
			assert len (kargs) == 1
			self.k, self.v = kargs.popitem ()
		elif len (args) == 2:
			self.k, self.v = args
		elif not args or args [0] is None:
			self.k, self.v = "", None
		else:
			self.k, self.v = "__eq", args [0]
		self._exclude = False

	def render (self, engine = DB_PGSQL):
		if self.v is None:
			return "NULL"
		_Q.render (self, engine)
		return self._value

	def __str__ (self):
		return self.render ()


def batch (**filters):
	Qs = []
	for k, v in filters.items ():
		if v is None:
			continue
		Qs.append (Q (k, v))
	return Qs



from . import utils
from .q import Q, V, batch, _Q
from .d import toval, D
from copy import deepcopy
from .dbtypes import DB_PGSQL, DB_SQLITE3
from .skitai_compat import EMPTY

class SQL:
    def __init__ (self, template, engine = DB_PGSQL, conn = None, check_filter = False):
        self._template = template
        self._engine = engine
        self._conn = conn
        self._check_filter = check_filter
        self._filters = []
        self._limit = 0
        self._offset = 0
        self._order_by = None
        self._group_by = None
        self._having = None
        self._returning = None
        self._insert_into = None
        self._cte = []
        self._feed = {}
        self._data = {}
        self._unionables = []
        self._statements = []
        self._transact = False
        self._explicit_empty = False

    def branch (self, conn = None):
        new = self.__class__ (self._template, self._engine, conn or self._conn)
        new._filters = deepcopy (self._filters)
        new._limit = deepcopy (self._limit)
        new._offset = deepcopy (self._offset)
        new._order_by = deepcopy (self._order_by)
        new._group_by = deepcopy (self._group_by)
        new._having = deepcopy (self._having)
        new._returning = deepcopy (self._returning)
        new._insert_into = deepcopy (self._insert_into)
        new._cte = deepcopy (self._cte)
        new._feed = deepcopy (self._feed)
        new._data = deepcopy (self._data)
        new._explicit_empty = self._explicit_empty
        new._check_filter = self._check_filter
        return new

    @property
    def query (self):
        return self.as_sql ()

    def addD (self, prefix, D_):
        assert prefix != "this", "Cannot use data prefix `this`"
        D_.encode (self._engine)
        self._data [prefix] = D_

    def render (self):
        return self.as_sql ()

    def __str__ (self):
        return self.as_sql ()

    def __getitem__(self, key):
        key.start and self.offset (key.start)
        if key.stop:
            self.limit (key.stop - (key.start or 0))
        return self

    def execute (self):
        if self._explicit_empty:
            return self
        if self._check_filter:
            assert self._filters, "No filter for modification, if you want to modify anyway use .all ()"
        return self._conn.execute (self.query) or self

    def all (self):
        self._filters.append ("1 = 1")
        return self

    def exclude (self, *Qs, **filters):
        g = []
        for q in Qs + tuple (batch (**filters)):
            if not q:
                continue
            if not isinstance (q, str):
                q.render (self._engine)
            g.append (str (q))

        cluses = " AND ".join (g)
        if cluses:
            self._filters.append ("NOT (" + cluses + ")")
        return self

    def into (self, table, *columns):
        self._insert_into = "INSERT INTO " + table
        if columns:
            self._insert_into += ' ({})'.format (', '.join (columns))
        return self

    def returning (self, *args):
        self._returning = "RETURNING " + ", ".join (args)
        return self

    def filter (self, *Qs, **filters):
        for q in Qs + tuple (batch (**filters)):
            if not q:
                continue
            if not isinstance (q, str):
                q.render (self._engine)
            self._filters.append (str (q))
        return self

    def having (self, cond):
        self._having = "HAVING " + cond
        return self

    def order_by (self, *by):
        self._order_by = utils.make_orders (by)
        return self

    def group_by (self, *by):
        self._group_by = utils.make_orders (by, "GROUP")
        return self

    def limit (self, val):
        if val == 0:
            self._explicit_empty = True
        self._limit = "LIMIT {}".format (val)
        return self

    def offset (self, val):
        self._offset = "OFFSET {}".format (val)
        return self

    def with_ (self, alias, sql):
        if isinstance (sql, D):
            self._cte.append ("{} ({}) AS (values ({}))".format (alias, sql.columns, sql.values))
        else:
            self._cte.append ("{} AS ({})".format (alias, str (sql)))
        return self

    def data (self, __donotusethisvariable__ = None, **karg):
        if __donotusethisvariable__:
            d = __donotusethisvariable__
            d.update (karg)
        else:
            d = karg

        for k, v in d.items ():
            if isinstance (v, D):
                self.addD (k, v)
            else:
                self._data [k] = toval (v, self._engine)
        return self

    def feed (self, **karg):
        for k, v in karg.items ():
            if isinstance (v, D):
                self.addD (k, v)
            else:
                # Q need str()
                if isinstance (v, _Q) and not v:
                    # for ignoring
                    v = "1 = 1"
                elif isinstance (v, (V, _Q)):
                    v.render (self._engine)
                self._feed [k] = str (v)
        return self

    def as_sql (self):
        raise NotImplementedError

    def tran (self):
        self._transact = True

    def append (self, sql):
        self._statements.append (sql)
        return self

    def intersect (self, sql):
        self._unionables.append ((sql, "INTERSECT"))
        return self

    def except_ (self, sql):
        self._unionables.append ((sql, "EXCEPT"))
        return self

    def union (self, sql):
        self._unionables.append ((sql, "UNION"))
        return self

    def union_all (self, sql):
        self._unionables.append ((sql, 'UNION ALL'))
        return self

    def maybe_union (self, current):
        n_q = len (self._unionables)
        qs = []
        qs.append (current)
        for idx, (sql, type) in enumerate (self._unionables):
            qs.append (type)
            qs.append (str (sql))
        r = "\n".join (qs)
        if self._transact:
            return "BEGIN TRANSACTION;\n" + r + ";\nCOMMIT;"
        return r

    # only work if self._explicit_empty ---------------------------
    def fetchall (self, *args, **kargs):
        return []

    def fetchmany (self, *args, **kargs):
        return []

    def fetchone (self, *args, **kargs):
        return

    def dispatch (self, *args, **kargs):
        return EMPTY

    def one (self, *args, **kargs):
        return EMPTY.one (*args, **kargs)

    def fetch (self, *args, **kargs):
        return EMPTY.fetch (*args, **kargs)


class TemplateParams:
    def __init__ (self, this, data):
        self.filter = " AND ".join ([f for f in this._filters if f])
        self.limit = this._limit
        self.offset = this._offset
        self.group_by = this._group_by
        self.order_by = this._order_by
        self.having = this._having
        self.returning = this._returning
        self.insert_into = this._insert_into
        self.cte = this._cte

        self.columns = data.columns
        self.values = data.values
        self.pairs = data.pairs

class SQLTemplateRenderer (SQL):
    def __call__ (self, **karg):
        return self.feed (**karg)

    def as_sql (self):
        data = utils.D (**self._data)
        this = TemplateParams (self, data)
        self._feed.update (self._data)
        self._feed ["this"] = this
        if self._template.find ("{_") != -1:
            compatables = {
                "_filters": this.filter,
                "_limit": this.limit,
                "_offset": this.offset,
                "_order_by": this.order_by,
                "_group_by": this.group_by,
                "_having": this.having,
                "_returning": this.returning,
                "_columns": data.columns,
                "_values": data.values,
                "_pairs": data.pairs
            }
            self._feed.update (compatables)
        r = self._template.format (**self._feed)
        return self.maybe_union (r)

class SQLComposer (SQL):
    def __init__ (self, template, engine = DB_PGSQL, conn = None, check_filter = False):
        SQL.__init__ (self, template, engine, conn, check_filter)
        self._joins = []

    def branch (self, conn = None):
        new = SQL.branch (self, conn)
        new._joins = deepcopy (self._joins)
        return new

    def get (self, *columns):
        self._feed ["select"] = ", ".join (columns)
        return self

    def _join (self, jtype, obj, alias, on, *Qs, **filters):
        if alias == 'true':
            alias, on = '',  'true'
        if alias:
            if alias.find (".") != -1:
                Qs = (alias,) + Qs
                alias = ''
            else:
                alias = ' AS {}'.format (alias)

        _filters = []
        for q in (on,) + Qs + tuple (batch (**filters)):
            if not q:
                continue
            if not isinstance (q, str):
                q.render (self._engine)
            _filters.append (str (q))
        _filters = " AND ".join ([f for f in _filters if f])
        if not isinstance (obj, str):
            # SQL
            obj = "({})".format (obj)
        self._joins.append (
            "{} {}{}{}".format (jtype, obj, alias, _filters and (' ON ' + _filters) or '')
        )
        return self

    def from_ (self, obj, alias = '', on = None, *Qs, **filters):
        return self._join ("FROM", obj, alias, on, *Qs, **filters)

    def join (self, obj, alias = '', on = None, *Qs, **filters):
        return self._join ("INNER JOIN", obj, alias, on, *Qs, **filters)

    def left_join (self, obj, alias = '', on = None, *Qs, **filters):
        return self._join ("LEFT OUTER JOIN", obj, alias, on, *Qs, **filters)

    def right_join (self, obj, alias = '', on = None, *Qs, **filters):
        return self._join ("RIGHT OUTER JOIN", obj, alias, on, *Qs, **filters)

    def full_join (self, obj, alias = '', on = None, *Qs, **filters):
        return self._join ("FULL OUTER JOIN", obj, alias, on, *Qs, **filters)

    def cross_join (self, obj, alias = '', on = None, *Qs, **filters):
        return self._join ("CROSS JOIN", obj, alias, *Qs, on, **filters)

    def as_sql (self):
        data = utils.D (**self._data)
        self._feed ["this"] = data
        if self._template.find ("{_") != -1:
            compatables = {
                "_columns": data.columns,
                "_values": data.values,
                "_pairs": data.pairs
            }
            self._feed.update (compatables)
        sql = [
            self._template.format (**self._feed)
        ]
        for join in self._joins:
            sql.append (join)
        _filters = [f for f in self._filters if f]
        _filters and sql.append ("WHERE " + " AND ".join (_filters))
        if self._group_by:
            sql.append (self._group_by)
            self._having and sql.append (self._having)
        self._order_by and sql.append (self._order_by)
        self._limit and sql.append (self._limit)
        self._offset and sql.append (self._offset)
        self._returning and sql.append (self._returning)
        if self._insert_into:
            sql.insert (0, self._insert_into)
        if self._cte:
            sql.insert (0, "WITH " + ", ".join (self._cte))
        r = "\n".join (sql)
        _ = self.maybe_union (r)
        if self._statements:
            _ += ";\n" + ";\n".join ([s.as_sql () for s in self._statements])
        return _



class F:
	def __init__ (self, val):
		self.val = val
		
	def __str__ (self):
		return self.val
	
	def add_percent (self, pos = 0):
		if pos == 0:
			self.val = "'%' || " + str (self.val) + " || '%'"
		elif pos == 1:	
			self.val = "'%' || " + str (self.val)
		else:
			self.val = str (self.val) + " || '%'"



import sqlite3
import json
import zlib
import sqlphile
from . import sql
from .dbtypes import DB_SQLITE3
from .skitai_compat import dispatch
from rs4 import asyncore
from rs4.cbutil import tuple_cb

class AttrDict (dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self
        
class open:
    def __init__ (self, path, dir = None, auto_reload = False, auto_closing = True):        
        self.closed = False
        self.auto_closing = auto_closing
        self.conn = None
        self.conn = sqlite3.connect (path, check_same_thread = False)        
                
        self.expt = None
        self._init (dir, auto_reload, DB_SQLITE3)   
    
    def _init (self, dir, auto_reload, engine):    
        self.create_cursor ()        
        self.sqlphile = sqlphile.SQLPhile (dir, auto_reload, engine = engine, conn = self)
    
    def create_cursor (self):    
        self.cursor = self.c = self.conn.cursor ()        
    
    def set_autocommit (self, flag = None):
       # flags: None (autocommit), DEFERRED, IMMEDIATE or EXCLUSIVE
       self.conn.isolation_level = flag
        
    def __enter__ (self):        
        return self
    
    def __exit__ (self, type, value, tb):        
        self.auto_closing and self.close ()

    def __del__ (self):
        self.close ()

    def __getattr__ (self, name):        
        try: 
            return getattr (self.c, name)
        except AttributeError:
            return getattr (self.sqlphile, name)

    def close (self):
        if self.conn and not self.closed:
            self.c.close ()
            self.conn.close ()
            self.closed = True

    def commit (self):    
        return self.conn.commit ()
    
    def rollback (self):    
        return self.conn.rollback ()    

    def serialize (self, obj):
        return zlib.compress (json.dumps (obj).encode ("utf8"))    
    
    def deserialize (self, data):
        return json.loads (zlib.decompress (data).decode ('utf8'))
    
    def blob (self, obj):
        return sqlite3.Binary (obj)    
    
    def field_names (self):
        return [x [0] for x in self.description]
        
    def as_dict (self, row, field_names = None):        
        return AttrDict (dict ([(f, row [i]) for i, f in enumerate (field_names or self.field_names ())]))
    
    def execute (self, sql, *args, **kargs):
        if isinstance (sql, (list, tuple)):
            sql = ";\n".join (map (str, sql)) + ";"              
        try:
            self.cursor.execute (str (sql), *args, **kargs)
        except:
            self.expt = asyncore.compact_traceback () [2]
            raise
        return self
    
    def fetchone (self, as_dict = False):
        try: 
            rows = self.fetchall (as_dict)
        except IndexError:
            return None    
        
    def fetchall (self, as_dict = False):
        return self.fetchmany (0, as_dict)
    
    def fetchmany (self, limit, as_dict = False):
        rows = limit and self.cursor.fetchmany (limit) or self.cursor.fetchall ()
        if not as_dict:
            return rows        
        field_names = self.field_names ()
        return [self.as_dict (row, field_names) for row in rows]
    
    def dispatch (self, *args, **kargs):
        return dispatch (self.fetchall (True), self.expt)

    def one (self, *args, **kargs):
        from skitai import exceptions
        
        rows = self.fetchall (True)
        if not rows:
            raise exceptions.HTTPError ("410 Partial Not Found")
        if len (rows) > 1:
            raise exceptions.HTTPError ("409 Conflict")    
        return rows [0]
    
    def fetch (self, *args, **kargs):
        return self.fetchall (True)
    

class open2 (open):
    def __init__ (self, conn, dir = None, auto_reload = False, auto_closing = True):
        self.closed = False
        self.conn = conn       
        self.auto_closing = auto_closing 
        self._init (dir, auto_reload, DB_SQLITE3)


# for faking skitai.was.db objects ---------------------------------------------
# single connection multi cursors

class open3 (open2):
    def __init__ (self, conn, dir = None, auto_reload = False):
        self.closed = False
        self.conn = conn   
        self.auto_closing = False
        self._init (dir, auto_reload, DB_SQLITE3)
        
    def _init (self, dir, auto_reload, engine):
        self.sqlphile = sqlphile.SQLPhile (dir, auto_reload, engine = engine, conn = self)
        self.cursor = self.c = None        
        self.cursors = []

    def __del__ (self):
        for each in self.cursors:
            each.close ()
        self.conn.close ()        

    def execute (self, *args, **kargs):
        cursor = self.conn.cursor ()
        self.cursors.append (cursor)
        return Executed (self, cursor, *args, **kargs)


class Executed (open):
    def __init__ (self, conn, cursor, *args, **kargs):
        self.conn = conn
        self.cursor = self.c = cursor        
        self.expt = None        
        self.meta = {}
        try:
            self.execute (*args, **kargs)
        except:            
            self.expt = asyncore.compact_traceback () [2] 
        else:    
            self.conn.commit ()           
    
    def __del__ (self):
        self.conn = None

    def set_callback (self, callback, reqid = None, timeout = 10):
        if reqid is not None:
            self.meta ["__reqid"] = reqid                        
        tuple_cb (self, callback)

    def maybe_reraise (self):
        self.conn = None # break back ref. for ensure call __del__
        if self.expt:
            raise self.expt

    def dispatch (self, *args, **kargs):
        self.conn = None
        return open.dispatch (self, *args, **kargs)

    def commit (self, *args, **kargs):
        self.maybe_reraise ()
    
    def fetch (self, *args, **kargs):
        self.maybe_reraise ()        
        return open.fetch (self, *args, **kargs)
    
    def one (self, *args, **kargs):
        self.maybe_reraise ()
        return open.one (self, *args, **kargs)

    # lower version compatable --------------------
    def wait (self, timeout = 10, *args, **karg):
        pass
    getswait = getwait = dispatch    
    wait_or_throw = commit



import datetime
import time
from .d import D
from .f import F
import re

def toD (dict, engine):
    d = D (**dict)
    d.encode (engine)
    return d

def make_orders (order_by, keyword = "ORDER"):
    if isinstance (order_by, str):
        order_by = [order_by]
    orders = []
    for f in order_by:
        if f.find ("-") != -1:
            orders.append (f.replace ("-", '') + " DESC")
        elif f.find ("+") != -1:
            orders.append (f.replace ("+", ''))
        else:
            orders.append (f)
    return "{} BY {}".format (keyword, ", ".join (orders))

def mkdatetime (unixtime):
    return datetime.datetime.utcfromtimestamp (unixtime)

ENDING = re.compile ("[.,\s]")
def omit (data, limit = 50):
    if len (data) < limit:
        return data
    m = ENDING.search (data, limit - 10)
    if m:
        data = data [:m.start ()].strip ()
        if len (data) <= (limit - 3):
            return data + "..."
    return data [:limit - 3].strip () + "..."

def as_dict (conn, row):
    fields = [x.name for x in conn.description]
    return dict ([(f, _row [i]) for i, f in enumerate (fields)])

def csv_row (row, line_delim = "", encoding = None):
    d = []
    for cell in row:
        if isinstance (cell, str):
            d.append ('"{}"'.format (cell.replace ('"', '\"')))
        elif isinstance (cell, datetime.datetime):
            d.append (cell.strftime ("%Y-%m-%d %H:%M:%S"))
        elif cell is None:
            d.append ('""')
        else:
            d.append (str (cell))
    row = "{}{}".format (",".join (d), line_delim)
    if encoding:
        row = row.encode (encoding)
    return row




"""
Hans Roh 2015 -- http://osp.skitai.com
License: BSD
"""
import re
import sys
import os
import shutil, glob
import codecs
from warnings import warn
try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup

with open('sqlphile/__init__.py', 'r') as fd:
	version = re.search(r'^__version__\s*=\s*"(.*?)"',fd.read(), re.M).group(1)

classifiers = [
  'License :: OSI Approved :: MIT License',
  'Development Status :: 4 - Beta',
  'Topic :: Database',
	'Intended Audience :: Developers',
	'Programming Language :: Python',
	'Programming Language :: Python :: 3',
]

packages = [
	'sqlphile',
]

package_dir = {'sqlphile': 'sqlphile'}
package_data = {}

install_requires = [
	"rs4"
]

with codecs.open ('README.rst', 'r', encoding='utf-8') as f:
	long_description = f.read()

setup(
	name='sqlphile',
	version=version,
	description='SQL Phile',
	long_description=long_description,
	url = 'https://gitlab.com/hansroh/sqlphile',
	author='Hans Roh',
	author_email='hansroh@gmail.com',
	packages=packages,
	package_dir=package_dir,
	package_data = package_data,
	license='MIT',
	platforms = ["posix", "nt"],
	download_url = "https://pypi.python.org/pypi/sqlphile",
	install_requires = install_requires,
	classifiers=classifiers
)



import datetime
from .f import F
from .dbtypes import DB_PGSQL, DB_SQLITE3

def toval (s, engine = DB_PGSQL):
	from .sql import SQLComposer

	if isinstance (s, datetime.datetime):
		if engine == DB_PGSQL:
			return "TIMESTAMP '" + s.strftime ("%Y-%m-%d %H:%M:%S") + "'"
		else:
			return "'" + s.strftime ("%Y-%m-%d %H:%M:%S") + "'"

	if isinstance (s, datetime.date):
		if engine == DB_PGSQL:
			return "TIMESTAMP '" + s.strftime ("%Y-%m-%d %H:%M:%S") + "'"
		else:
			return "'" + s.strftime ("%Y-%m-%d %H:%M:%S") + "'"

	if s is None:
		return "NULL"

	if isinstance (s, bool):
		return s == True and 'true' or 'false'

	if isinstance (s, (SQLComposer,)):
		return "({})".format (str (s))

	if isinstance (s, (float, int, F)):
		return str (s)

	return "'" + s.replace ("'", "''") + "'"

class D:
	def __init__ (self, **data):
		self._feed = data
		self._columns = list (self._feed.keys ())
		self._encoded = False

	def encode (self, engine = DB_PGSQL):
		if self._encoded:
			return
		_data = {}
		for k, v in self._feed.items ():
			_data [k] = toval (v, engine)
		self._feed = _data
		self._encoded = True
		return self

	@property
	def columns (self):
		return ", ".join (self._columns)

	@property
	def values (self):
		return ", ".join ([self._feed [c] for c in self._columns])

	@property
	def pairs (self):
		return ", ".join (["{} = {}".format (c, self._feed [c]) for c in self._columns])
